
(*
	----------------------------------------------

	  recorderApp.pas
	  Voice Communicator II
	  Audio Tools - simple wave recorder application class

	----------------------------------------------
	  This source code cannot be used without
	  proper permission granted to you as a private
	  person or an entity by the Lake of Soft, Ltd

	  Visit http://lakeofsoft.com/ for details.

	  Copyright (c) 2001, 2002 Lake of Soft, Ltd
		     All rights reserved
	----------------------------------------------

	  created by:
		Lake, 01 Nov 2001

	  modified by:
		Lake, 25 Jan 2002
		Lake, 04 Feb 2002
		Lake, 21 Feb 2002
		Lake, 13 Mar 2002
		Lake, 21 Mar 2002
		Lake, 25 Mar 2002
		Lake, 12 Apr 2002

	----------------------------------------------
*)

{$I unaDef.inc}

unit
  recorderApp;

interface

uses
  Windows, MMSystem,
  unaTypes, unaUtils, unaClasses, unavcApp,
  unaMsAcmClasses;

type
  unaRecorderApp = class (unaVCApplication)
  private
    f_total: int64;
  protected
    function init(): bool; override;
    procedure feedback(); override;
    function stop(): bool; override;
  public
  end;


implementation


{ unaRecorderApp }

// --  --
procedure unaRecorderApp.feedback();
var
  str: string;
begin
  inherited;
  //
  str := 'Recorder so far: ' + int2Str(device.outBytes, 10, 3) + #13;
{$IFDEF CONSOLE}
  write(str);
{$ELSE}
  infoMessage(str);
{$ENDIF}
end;

// --  --
function unaRecorderApp.init(): bool;
var
  out_file: string;
  append: bool;
begin
  result := inherited init();
  //
  if (result) then begin
    ini.setSection('data');
    out_file := ini.get('out_file', 'rec_buf.dat');
    append := ini.get('append', true);

    if ((1 > ParamCount()) or hasSwitch('h')) then
      infoMessage('Usage: [out_file] [rate [bits [nChannels]]] [/choose] [/h]'#13#10)
    else begin
      out_file := ParamStr(1);
      if (1 < ParamCount) then
	rate := str2Int(ParamStr(2), unsigned(44100));
      if (2 < ParamCount) then
	bits := str2Int(ParamStr(3), unsigned(16));
      if (3 < ParamCount) then
	nChannels := str2Int(ParamStr(4), unsigned(2));
    end;

    device := unaWaveInDevice.create(deviceId);
    assignFormat();
    if (not mmNoError(device.open())) then
      unaWaveDevice(device).direct := false
    else
      device.close();

    // assign output stream
    with (unaFileStream(device.assignStream(false, unaFileStream))) do begin
      initStream(out_file, GENERIC_WRITE);
      if (append) then
	seek(0, false)
      else
	clear();
    end;

{$IFDEF VC20_SU_LICENSE}
{$ELSE}
    assignVolumeParams();
{$ENDIF}

    infoMessage('--------------------------');
    infoMessage('Input device     : ' + unaWaveInDevice(device).getCaps().szPname + choice(unaWaveDevice(device).direct, ' [direct]', ''));
    infoMessage(' ');
{$IFDEF CONSOLE}
    infoMessage('Sampling rate       : ' + int2Str(rate));
    infoMessage('Bits per sample     : ' + int2Str(bits));
    infoMessage('Number of channels  : ' + int2Str(nChannels));
    assert(debugMessage('Chunk size          : ' + int2Str(device.chunkSize)));
    infoMessage(' ');
{$ENDIF}

{$IFDEF VC20_SU_LICENSE}
{$ELSE}
    infoMessage('Silence detection level : ' + choice(0 < unaWaveInDevice(device).minVolumeLevel, int2Str(unaWaveInDevice(device).minVolumeLevel), 'disabled'));
{$ENDIF}
    infoMessage('Output file name : ' + out_file);
    infoMessage('Create new file  : ' + choice(not append, 'Yes', 'No'));
    infoMessage('--------------------------'#13#10);

{$IFNDEF CONSOLE}
    if (0 < ParamCount) then
      doStart();
{$ENDIF}
  end;
end;

// --  --
function unaRecorderApp.stop(): bool;
begin
  result := inherited stop();
  //
  if (result) then begin
    inc(f_total, device.outBytes);
    infoMessage('Total bytes recorded: ' + int2Str(f_total, 10, 3));
  end;
end;

end.

