
(*
	----------------------------------------------

	  vcConsoleApp.pas
	  Voice Communicator II
	  Audio Tools - console demonstration application

	----------------------------------------------
	  This source code cannot be used without
	  proper permission granted to you as a private
	  person or an entity by the Lake of Soft, Ltd

	  Visit http://lakeofsoft.com/ for details.

	  Copyright (c) 2001, 2002 Lake of Soft, Ltd
		     All rights reserved
	----------------------------------------------

	  created by:
		Lake, 25 Mar 2002

	  modified by:
		Lake, 25 Mar 2002
		Lake, 12 Apr 2002

	----------------------------------------------
*)

{$I unaDef.inc}

unit
  vcConsoleApp;

interface

uses
  Windows, unaTypes,
  unaUtils, unaClasses, WinSock, unaSockets, unavcApp,
  unaMsAcmClasses;

const
  // version
  vcConsole_minversion	= $10;
  vcConsole_curversion	= $10;

  // codes
  vcConsole_Code_ping	= $00;
  vcConsole_Code_pong	= $01;
  //
  vcConsole_Code_askConnection		= $10;
  vcConsole_Code_grantConnection	= $11;
  vcConsole_Code_refuseConnection	= $12;
  //
  vcConsole_Code_askDisconnection	= $15;
  vcConsole_Code_grantDisconnection	= $16;
  //
  vcConsole_Code_RTDPData		= $20;

type

  //
  // -- vcConsolePacket --
  //

  pvcConsolePacket = ^vcConsolePacket;
  vcConsolePacket = packed record
    r_version: byte;
    r_code: byte;
    r_seq: unsigned;
    r_dataSize: word;
    r_data: record end;
  end;


  //
  // -- vcConsolePacketData_StreamFormat --
  //

  pvcConsolePacketData_StreamFormat = ^vcConsolePacketData_StreamFormat;
  vcConsolePacketData_StreamFormat = packed record
    r_mid: word;
    r_pid: word;
    r_srcTag: unsigned;
    r_srcIndex: unsigned;
    r_dstTag: unsigned;
    r_dstIndex: unsigned;
  end;


  //
  // -- vcConsolePeer --
  //

  unavcConsoleApp = class;

  vcConsolePeer = class(unaThread)
  private
    f_state: unsigned;
    f_connId: unsigned;
    f_lastLocalSeq: unsigned;
    f_lastRemoteSeq: unsigned;
    f_lastRemotePong: unsigned;
    f_joinReqCount: unsigned;
    f_pingReqCount: unsigned;
    f_pingSent: unsigned;
    //
    f_stat_inUDPBytes: unsigned;
    f_stat_outUDPBytes: unsigned;
    f_stat_skipPlayBytes: unsigned;
    //
    f_master: unavcConsoleApp;
    f_codec: unaMsAcmCodec;
    f_dataEvent: unaEvent;
    f_timer: unaThreadTimer;
    //
    procedure onTimer(ticks: unsigned);
    procedure onData(data: pointer; size: unsigned);
    procedure sendPacket(code: byte; data: pointer = nil; size: unsigned = 0);
    function waitForState(value: unsigned): bool;
    procedure close(inDir: bool);
    procedure open(inDir: bool);
  protected
    function grantStop(): bool; override;
    procedure startIn(); override;
    procedure startOut(); override;
    function execute(): int; override;
  public
    constructor create(master: unavcConsoleApp; connId: unsigned);
    destructor destroy; override;
    //
  end;


  //
  // -- unavcConsoleApp --
  //

  unavcConsoleApp = class(unaVCApplication)
  private
    f_host: string;
    f_port: word;
    f_socksId: unsigned;
    f_streamFormat: vcConsolePacketData_StreamFormat;
    //
    f_acm: unaMsAcm;
    f_socks: unaSocks;
    f_player: unaWaveOutDevice;
    f_recorder: unaWaveInDevice;
    f_localCodec: unaMsAcmCodec;
    f_isServer: bool;
    f_peer: vcConsolePeer;
    //
    procedure onSockEvent(event: unaSocketEvent; id, connId: unsigned; data: pointer; size: unsigned);
  protected
    function init(): bool; override;
    function start(): bool; override;
    function stop(): bool; override;
  public
    destructor destroy(); override;
  end;


implementation


uses
  MMSystem;

{ vcConsolePeer }

// --  --
procedure vcConsolePeer.close(inDir: bool);
begin
  //
  if (inDir) then begin
    f_codec.free();
    f_codec := nil;
    f_master.f_player.close();
  end
  else begin
    f_master.f_recorder.close();
    f_master.f_localCodec.close();
  end;
end;

// --  --
constructor vcConsolePeer.create(master: unavcConsoleApp; connId: unsigned);
begin
  inherited create();
  //
  f_master := master;
  f_connId := connId;
  f_dataEvent := unaEvent.create();
  f_timer := unaThreadTimer.create(400);
  f_timer.onTimer := onTimer;
end;

// --  --
destructor vcConsolePeer.destroy();
begin
  inherited;
  //
  f_codec.free();
  f_dataEvent.free();
  f_timer.free();
end;

// --  --
function vcConsolePeer.execute(): int;
var
  bufRec: pointer;
  bufRecSize: unsigned;
//  bufRecLoad: unsigned;
  //
  bufCodec: pointer;
  bufCodecSize: unsigned;
  bufCodecLoad: unsigned;
begin
  bufRecSize := f_master.f_recorder.chunkSize;
  getMem(bufRec, bufRecSize);
  bufCodecSize := 1800;//f_master.f_localCodec.chunkSize;
  getMem(bufCodec, bufCodecSize);
  //
  while (not shouldStop) do begin
    //
    f_dataEvent.waitFor();

    if (enter(10)) then try

      case (f_state) of

	0: begin	// not in chat
	  // first, ping the peer
	  inc(f_pingReqCount);
	  //
	  if (4 >= f_pingReqCount) then begin
	    f_pingSent := GetTickCount();
	    f_state := 1;
	    debugMessage(' * waiting for peer pong (' + int2Str(f_pingReqCount) + '/4)');
	    sendPacket(vcConsole_Code_ping);
	  end
	  else begin
	    f_state := 2;	// no peer
	    debugMessage(' * peer was not found');
	  end;
	end;

	1: begin	// waiting for first pong
	  if (10000 < getTickCount() - f_pingSent) then begin	// more that 10 sec
	    f_state := 0;	// send again
	  end
	end;

	2: begin	// no peer
	  askStop();
	end;

	3: begin	// peer was found
	  // ask stream
	  inc(f_joinReqCount);
	  if (4 >= f_joinReqCount) then begin
	    f_pingReqCount := 0;
	    f_pingSent := GetTickCount();
	    f_state := 4;
	    debugMessage(' * waiting for peer grant to join (' + int2Str(f_joinReqCount) + '/4)');
	    sendPacket(vcConsole_Code_askConnection, @f_master.f_streamFormat, sizeOf(f_master.f_streamFormat));
	  end
	  else begin
	    debugMessage(' * unable to join due to constant timeout from peer');
	    f_state := 5;	// not granted
	  end
	end;

	4: begin	// waiting fot grant or refuse
	  if (10000 < getTickCount() - f_pingSent) then begin
	    // unable to wait for grant, pinging again..
	    f_state := 0;
	    debugMessage(' * unable to join due to timeout from peer');
	  end;
	end;

	5: begin	// unable to join or want to leave
	  close(false);
	  close(true);
	  //
	  sendPacket(vcConsole_Code_askDisconnection);
	  f_state := 2;
	end;

	6: begin	// granted
	  {
	  // 1. read recorder
	  if (f_master.f_recorder.okToRead()) then
	    bufRecLoad := f_master.f_recorder.read(bufRec, bufRecSize)
	  else
	    bufRecLoad := 0;

	  // 2. feed codec
	  if (0 < bufRecLoad) then
	    if (f_master.f_localCodec.isOpen()) then
	      f_master.f_localCodec.write(bufRec, bufRecLoad);
	  }

	  // 3. check codec output
	  if (f_master.f_localCodec.okToRead()) then
	    bufCodecLoad := f_master.f_localCodec.read(bufCodec)
	  else
	    bufCodecLoad := 0;

	  // 4. put codec output into socket
	  if (0 < bufCodecLoad) then begin
	    sendPacket(vcConsole_Code_RTDPData, bufCodec, bufCodecLoad);
	    //
	    inc(f_stat_outUDPBytes, bufCodecLoad);
	  end;

	  // 5. check own codec output
	  {
	  bufCodecLoad := 0;
	  if (nil <> f_codec) then
	    if (f_codec.okToRead()) then
	      bufCodecLoad := f_codec.read(bufCodec, bufCodecSize);

	  // 6. put data from own codec into player
	  if (0 < bufCodecLoad) then
	    if (f_master.f_player.isOpen()) then
	      if (f_master.f_player.okToWrite()) then
		f_master.f_player.write(bufCodec, bufCodecLoad)
	      else
		inc(f_stat_skipPlayBytes, bufCodecLoad);
	  }
	end;

	7:
	  askStop();

      end;

    finally
      leave();
    end;

  end;
  //
  freeMem(bufCodec);
  freeMem(bufRec);
  result := 0;
end;

// --  --
function vcConsolePeer.grantStop(): bool;
begin
  case (f_state) of
    4, 6: begin
      close(false);
      close(true);
      //
      sendPacket(vcConsole_Code_askDisconnection);
      f_state := 2;
    end  
    else
      f_state := 2;
  end;
  //
  result := waitForState(2);
end;

// --  --
procedure vcConsolePeer.onData(data: pointer; size: unsigned);
var
  packet: pvcConsolePacket;
  packetData: pointer;
  streamFormat: pvcConsolePacketData_StreamFormat;
  driver: unaMsAcmDriver;
begin
  if (size >= sizeof(packet^)) then begin
    packet := data;
    if (
	 (size = sizeof(packet^) + packet.r_dataSize) and
	 (packet.r_version >= vcConsole_minversion) and
	 (packet.r_seq >= f_lastRemoteSeq)
       ) then begin

      // looks like a valid packet
      packetData := @packet.r_data;
      if (f_lastRemoteSeq + 1 <> packet.r_seq) then
	assert(debugMessage(' * skipped: ' + int2Str(packet.r_seq - f_lastRemoteSeq - 1) + ' packets'));
      f_lastRemoteSeq := packet.r_seq;
      case (packet.r_code) of

	//
	vcConsole_Code_ping: begin
	  // send pong
	  sendPacket(vcConsole_Code_pong);
	end;

	//
	vcConsole_Code_pong: begin
	  // mark time
	  f_lastRemotePong := getTickCount();
	  if (1 = f_state) then begin
	    enter(100);
	    try
	      f_state := 3;	// connected
	    finally
	      leave();
	    end;
	  end;
	end;

	//
	vcConsole_Code_askConnection: begin
	  // check if we can accept connection stream
	  if (packet.r_dataSize >= sizeOf(streamFormat^)) then begin
	    streamFormat := @packet.r_data;
	    enter(1000);
	    try
	      f_codec.free();
	      f_master.f_player.close();
	      //
	      driver := f_master.f_acm.getDriver(streamFormat.r_mid, streamFormat.r_pid);
	      f_codec := unaMsAcmCodec.create(driver, false);
	      f_codec.setFormat(true, streamFormat.r_dstTag, streamFormat.r_dstIndex);
	      f_codec.setFormat(false, streamFormat.r_srcTag, streamFormat.r_srcIndex);
	      //
	      f_master.f_player.setSampling(f_codec.dstFormat.nSamplesPerSec, f_codec.dstFormat.wBitsPerSample, f_codec.dstFormat.nChannels);
	      f_codec.addConsumer(f_master.f_player);
	      //
	      if (mmNoError(f_codec.open())) then begin
		f_master.f_player.open();
		sendPacket(vcConsole_Code_grantConnection);
		infoMessage('-------------------------------');
		debugMessage(' * we accept the following remote stream:');
		debugMessage('-------------------------------');
		infoMessage('remote codec driver : ' + driver.getDetails().szShortName);
		infoMessage('in-format           : ' + f_codec.srcFormatInfo);
		infoMessage('out-format (PCM)    : ' + f_codec.dstFormatInfo);
		infoMessage('chunk size          : ' + int2Str(f_codec.chunkSize));
		infoMessage('status              : active');
		infoMessage('-------------------------------');
	      end
	      else begin
		f_codec.free();
		f_codec := nil;
		debugMessage(' * unable to accept remote peer stream');
		sendPacket(vcConsole_Code_refuseConnection);
	      end;
	    finally
	      leave();
	    end;
	  end;
	end;

	//
	vcConsole_Code_grantConnection: begin
	  // connection was granted to us
	  if (4 = f_state) then begin
	    debugMessage(' * our audio stream was accpeted, activating devices now..');
	    enter(1000);
	    try
	      open(false);
	      f_state := 6;	// granted
	      debugMessage(' * activated, you can start speaking now');
	    finally
	      leave();
	    end;
	  end;
	end;

	//
	vcConsole_Code_askDisconnection: begin
	  enter(400);
	  try
	    close(true);
	    f_state := 0;
	  finally
	    leave();
	  end;
	  debugMessage(' * remote peer was disconnected');
	  sendPacket(vcConsole_Code_grantDisconnection);
	  close(false);
	  //
	  f_state := 7;
	end;

	//
	vcConsole_Code_grantDisconnection: begin
	  close(false);
	  f_state := 2;
	  debugMessage(' * disconnected gracefully');
	end;

	//
	vcConsole_Code_RTDPData: begin
	  if ((6 = f_state) and (nil <> f_codec)) then begin
	    // put packet data into codec
	    if (f_codec.isOpen()) then
	      f_codec.write(packetData, packet.r_dataSize);
	    //
	    inc(f_stat_inUDPBytes, packet.r_dataSize);
	  end
	  else
	    assert(debugMessage(' * RTDP packet ignored'));
	end;

      end;
    end
    else
      assert(debugMessage(' * invalid packet skipped'));
  end;
end;

// --  --
procedure vcConsolePeer.onTimer(ticks: unsigned);
begin
{$IFDEF CONSOLE }
  write('[REC: ' + int2Str(f_master.f_recorder.outBytes, 10, 3) + '] ' +
	'[UDP Out: ' + int2Str(f_stat_outUDPBytes, 10, 3) + '] ' +
	'[UDP In: ' + int2Str(f_stat_inUDPBytes, 10, 3) + '] ' +
	'[PLAY: ' + int2Str(f_master.f_player.inBytes, 10, 3) + '] ' +
	'[SKIP:' + int2Str(f_stat_skipPlayBytes, 10, 3) + ']' +
	#13);
{$ELSE}
  infoMessage('[REC: ' + int2Str(f_master.f_recorder.outBytes, 10, 3) + '] ' +
	'[UDP Out: ' + int2Str(f_stat_outUDPBytes, 10, 3) + '] ' +
	'[UDP In: ' + int2Str(f_stat_inUDPBytes, 10, 3) + '] ' +
	'[PLAY: ' + int2Str(f_master.f_player.inBytes, 10, 3) + '] ' +
	'[SKIP:' + int2Str(f_stat_skipPlayBytes, 10, 3) + ']' +
	#13);
{$ENDIF}
end;

// --  --
procedure vcConsolePeer.open(inDir: bool);
begin
  if (inDir) then begin
    f_master.f_player.open();
    // f_codec should be already open
    f_stat_inUDPBytes    := 0;
    f_stat_skipPlayBytes := 0;
  end
  else begin
    f_master.f_localCodec.open();
    f_master.f_recorder.open();
    //
    f_stat_outUDPBytes   := 0;
  end;
end;

// --  --
procedure vcConsolePeer.sendPacket(code: byte; data: pointer; size: unsigned);
var
  packet: pvcConsolePacket;
begin
  if (not shouldStop) then begin
    getMem(packet, sizeOf(packet^) + size);
    //
    packet.r_version := vcConsole_curversion;
    packet.r_code := code;
    inc(f_lastLocalSeq);
    packet.r_seq := f_lastLocalSeq;
    packet.r_dataSize := size;
    if (size > 0) then
      move(data^, packet.r_data, size);

    // - emulate packets loss
    {
    if (20 = random(50)) then
      debugMessage(' * random packet lost')
    else}
      f_master.f_socks.sendData(f_master.f_socksId, packet, sizeof(packet^) + size, f_connId);

    freeMem(packet);
  end;
end;

// --  --
procedure vcConsolePeer.startIn();
begin
  inherited;
  //
  f_timer.start();
end;

// --  --
procedure vcConsolePeer.startOut();
begin
  inherited;
  //
  f_timer.stop();
end;

// --  --
function vcConsolePeer.waitForState(value: unsigned): bool;
var
  i: int;
begin
  i := 0;
  while (f_state <> value) do begin
    windows.Sleep(100);
    inc(i);
    if (4 < i) then
      break;
  end;
  //
  result := (f_state = value);
end;

{ unavcConsoleApp }

// --  --
destructor unavcConsoleApp.destroy();
begin
  inherited;
  //
  f_recorder.free();
  f_player.free();
  f_localCodec.free();
  f_socks.free();
  f_acm.free();
end;

// --  --
function unavcConsoleApp.init(): bool;
var
  res: MMRESULT;

  // player
  outDeviceId: unsigned;
  outStatus: string;

  // recorder
  inDeviceId: unsigned;
  inStatus: string;

  // codec
  driver: unaMsAcmDriver;
  codecStatus: string;
begin
  result := inherited init();
  if (result) then begin
{$IFDEF CONSOLE}
{$ELSE}
    formatLabel.show(SW_HIDE);
    chooseButton.show(SW_HIDE);
{$ENDIF}
    //
    f_socks := unaSocks.create();
    f_socks.onEvent := onSockEvent;

    // socket initialization
    ini.setSection('socket.settings');
    f_host := ini.get('host', '127.0.0.1');
    f_port := ini.get('port', unsigned(17600));

    // codec initialization
    ini.setSection('codec.driver');
    f_streamFormat.r_mid := ini.get('mid', unsigned(1));
    f_streamFormat.r_pid := ini.get('pid', unsigned(36));

    ini.setSection('codec.format');
    f_streamFormat.r_srcTag := ini.get('sourceTag', unsigned(1));
    f_streamFormat.r_srcIndex := ini.get('sourceIndex', unsigned(3));

    ini.setSection('codec.format');
    f_streamFormat.r_dstTag := ini.get('dstTag', unsigned(49));
    f_streamFormat.r_dstIndex := ini.get('dstIndex', unsigned(1));

    // codec & recorder initialization
    ini.setSection('recorder.device');
    inDeviceId := unsigned(ini.get('id', int(WAVE_MAPPER)));
    f_recorder := unaWaveInDevice.create(inDeviceId, false);

    f_acm := unaMsAcm.create();
    f_acm.enumDrivers();
    driver := f_acm.getDriver(f_streamFormat.r_mid, f_streamFormat.r_pid);
    if (nil <> driver) then begin
      f_localCodec := unaMsAcmCodec.create(driver, false, 5);
      //
      f_localCodec.setFormat(true, f_streamFormat.r_srcTag, f_streamFormat.r_srcIndex);
      f_localCodec.setFormat(false, f_streamFormat.r_dstTag, f_streamFormat.r_dstIndex);
      res := f_localCodec.open();
      if (mmNoError(res)) then begin
	codecStatus := 'ready';
	f_localCodec.close();
	//
	if (driver.assignFormat(f_streamFormat.r_srcTag, f_streamFormat.r_srcIndex, f_recorder)) then
	else

      end
      else begin
	codecStatus := 'error code: ' + int2Str(res);
	result := false;
      end;
    end
    else begin
      infoMessage('Unable to locate specified codec driver, mid=' + int2Str(f_streamFormat.r_mid) + ', pid=' + int2Str(f_streamFormat.r_pid));
      result := false;
    end;

    // --  --
    if (result) then begin
      //
      if (not mmNoError(f_recorder.open())) then begin
	f_recorder.direct := false;
      end
      else
	f_recorder.close();
      //
      res := f_recorder.open();
      if (mmNoError(res)) then begin
	f_recorder.close();
	inStatus := 'ready';
      end
      else
	inStatus := f_recorder.getErrorText(res);

      // player initialization

      ini.setSection('player.device');
      outDeviceId := unsigned(ini.get('id', int(WAVE_MAPPER)));

      f_player := unaWaveOutDevice.create(outDeviceId, false);
      f_player.setSampling(8000);	// use default, just for testing
      if (not mmNoError(f_player.open())) then begin
	f_player.direct := false;
      end
      else
	f_player.close();
      //
      f_recorder.open();
      res := f_player.open();
      if (mmNoError(res)) then begin
	f_player.close();
	outStatus := 'ready';
      end
      else
	outStatus := f_player.getErrorText(res);
      //
      f_recorder.close();
      f_recorder.addConsumer(f_localCodec);

      f_isServer := hasSwitch('server') or hasSwitch('srv');

      infoMessage(#13#10'--- vcConsole ------------------------');
      infoMessage(' ');
      infoMessage('mode             : ' + choice(f_isServer, 'server', 'client'));
      if (not f_isServer) then begin
	infoMessage('remote host      : ' + f_host);
	infoMessage('remote port      : ' + int2Str(f_port));
      end
      else
	infoMessage('local port       : ' + int2Str(f_port));
      infoMessage(' ');
      infoMessage('playback device  : ' + f_player.getCaps().szPname + choice(f_player.direct, ' [direct]', ''));
      assert(debugMessage('chunk size       : ' + int2Str(f_player.chunkSize)));
      infoMessage('status           : ' + outStatus);
      infoMessage(' ');
      infoMessage('recording device    : ' + f_recorder.getCaps().szPname + choice(f_recorder.direct, ' [direct]', ''));
      infoMessage('Sampling rate       : ' + int2Str(f_recorder.dstFormat.nSamplesPerSec));
      infoMessage('Bits per sample     : ' + int2Str(f_recorder.dstFormat.wBitsPerSample));
      infoMessage('Number of channels  : ' + bool2StrStr(1 <> f_recorder.dstFormat.nChannels));
      assert(debugMessage('chunk size          : ' + int2Str(f_recorder.chunkSize)));
      infoMessage('status              : ' + inStatus);
      infoMessage(' ');
      infoMessage('local codec driver : ' + driver.getDetails().szShortName);
      infoMessage('in-format (PCM)    : ' + f_localCodec.srcFormatInfo);
      infoMessage('out-format         : ' + f_localCodec.dstFormatInfo);
      assert(debugMessage('chunk size         : ' + int2Str(f_localCodec.chunkSize)));
      infoMessage('status             : ' + codecStatus);
      infoMessage(' ');
      infoMessage('--- --------- ------------------------'#13#10);
    end;
  end;

{$IFNDEF CONSOLE}
  if (result and (0 < ParamCount)) then
    doStart();
{$ENDIF}
end;

// --  --
procedure unavcConsoleApp.onSockEvent(event: unaSocketEvent; id, connId: unsigned; data: pointer; size: unsigned);
begin
  case (event) of
    unaseServerConnect,
    unaseClientConnect: begin
      if (nil <> f_peer) then
	// ignore all other connects
      else begin
	f_peer := vcConsolePeer.create(self, connId);
	if (f_isServer) then
	  debugMessage(' * remote peer connected')
	else
	  debugMessage(' * connection to remote server established');
	//
	f_peer.start();
      end;
    end;

    unaseServerData,
    unaseClientData: begin
      if ((nil <> f_peer) and (f_peer.f_connId = connId)) then
	f_peer.onData(data, size)
      else
	// ignore all other data
    end;

    unaseServerDisconnect,
    unaseClientDisconnect: begin
      // never happen on UDP, but just in case
      if ((nil <> f_peer) and (f_peer.f_connId = connId)) then begin
	f_peer.free();
	f_peer := nil;
	if (f_isServer) then
	  debugMessage(' * remote peer disconnected')
	else
	  debugMessage(' * connection to remote server has been closed');
      end
      else
	// ignore all other disconnects
    end;

    unaseThreadStartupError: begin
      // fatal error
      if (f_isServer) then
	debugMessage('unable to start listening on port ' + int2Str(f_port))
      else
	debugMessage('unable to ');
    end;

  end;
end;

// --  --
function unavcConsoleApp.start(): bool;
begin
  if (f_isServer) then
    f_socksId := f_socks.createServer(f_port, IPPROTO_UDP)
  else
    f_socksId := f_socks.createConnection(f_host, f_port, IPPROTO_UDP);
  //
  result := (0 <> f_socksId);
  if (f_isServer) then
    infoMessage(choice(result, 'server has been activated on port ' + int2Str(f_port), 'unable to activate server on port ' + int2Str(f_port)))
  else
    infoMessage(choice(result, 'client has been connected to ' + f_host + ':' + int2Str(f_port), 'unable to connect to ' + f_host + ':' + int2Str(f_port)));
  //
  if (result) then
    result := inherited start();
end;

// --  --
function unavcConsoleApp.stop(): bool;
begin
  result := inherited stop();
  if (result) then begin
    //
    f_peer.free();
    f_peer := nil;
    f_socks.delete(f_socksId);
    infoMessage(choice(f_isServer, 'server', 'client') + ' has been stopped.');
  end;  
end;

end.

